import torch
import regex
import numpy as np
import os
import time
import itertools
import numpy as np
import collections
from .._utils import Trie
from ._model import Model

try:
    from llama_cpp.llama import Llama, LlamaGrammar
except:
    pass

class MatchMock():
    __slots__ = ('partial')
    def groupdict(self):
        return self
PARTIAL_MATCH = MatchMock()
PARTIAL_MATCH.partial = True

METACHARS = frozenset("()[]{}?*+|^$\\.-#&~")
def const_prefix(pattern):
    for i,c in enumerate(pattern):
        if c in METACHARS:
            return i
    return len(pattern)



class LlamaCPP(Model):
    def __init__(self, model_path, echo=True):
        super().__init__(echo)

        if isinstance(model_path, str):
            self.model = Llama(model_path)

    def __call__(self, pattern=, max_tokens=100, n=1, top_p=1, temperature=0.0, ensure_bos_token=True):
        prompt = str(self)

        # add the beginning of sequence token if needed
        if ensure_bos_token and not prompt.startswith(self.bos_token):
            prompt = self.bos_token + prompt

        # find out how much of the prompt we have in the KV cache (and strip from the prompt what we already have)
        cache_token_ids = self._cache_state["cache_token_ids"]
        if len(cache_token_ids) > 0:
            cache_pos = 0
            for i,id in enumerate(cache_token_ids):
                token = self.tokens[id]
                if prompt[cache_pos:cache_pos+len(token)] != token:
                    self._cache_state["cache_token_ids"] = cache_token_ids[:i]
                    break
                cache_pos += len(token)
            # self._cache_tokens = self._cache_tokens[:i]
            prompt = prompt[cache_pos:]
        self._cache_state["new_token_ids"] = []

        # send the known prompt tokens to the KV cache as a batch
        forced_token_ids = []
        valid_id_set = None
        while True:
            token, token_id = self._longest_token_match(prompt)
            if token_id is not None:
                forced_token_ids.append(token_id)
                prompt = prompt[len(token):]
            else:
                if len(prompt) > 0:

                    # get all possible extensions
                    valid_id_set = self._token_trie.values(prompt)

                    # if make sure we also allow for the shortest maximal extension (even if that does not consume all the remaining text)
                    tnode = self._token_trie[prompt]
                    if tnode.value is None:
                        while tnode.value is None:
                            tnode = tnode.parent
                        valid_id_set.append(tnode.value)
                    valid_id_set = set(valid_id_set)
                break
        if len(forced_token_ids) > 0:
            self._cache_state["new_token_ids"].extend(forced_token_ids)
        
        # move whatever is not cached from the prompt into the pattern (since the pattern is what we will generate)
        # note we also anchor the pattern to the start of the sequence
        pattern = "^" + regex.escape(prompt, literal_spaces=True) + pattern
        if "(?P<stop>" not in pattern:
            pattern += "(?P<stop>" + regex.escape(self.eos_token) + ")"
        pattern_obj = regex.compile(pattern, flags=regex.DOTALL)
        const_prefix_len = const_prefix(pattern[1:])

        assert n == 1, "Still need to add support for n > 1!"

        extracted_stop_pattern = regex.compile("(" + pattern[pattern.index("(?P<stop>")+9:-1] + ")$", flags=regex.DOTALL)

        # loop until we have generated a complete pattern
        # call_count = [0]
        hidden_count = len(prompt) # we don't emit the part of the prompt we have to regenerate for token healing
        generated_text = ""
        delayed_text = ""
        sampled_token_ind = None
        for token_count in range(max_tokens):

            # TODO: eventually we could try and check if the regex "forces" the next token so we can skip
            #       logit computation entirely. This might only make sense if we can make the regex matching
            #       really fast (integrate with the FSM directly) or make it report when a character is forced.

            # here we check if there is only one possible token match, in which case we can skip computing the logits
            # TODO: [SML] this is WRONG, because just because there is only one possible token extension does not mean
            #       we might not have more options if we were allowed to make a new token. the way to fix this would be
            #       to integrate with the regex module so partial matches come with a return value that tells us if what
            #       comes next must be a specific character...
            found_node = None
            node_stack = [(generated_text + delayed_text, '', self._token_trie)]
            self._token_trie.match_version += 1
            match_version = self._token_trie.match_version
            gen_len = len(generated_text) + len(delayed_text)
            while len(node_stack) > 0:
                curr_text, curr_char, node = node_stack.pop()
                if node.match_version < match_version:
                    node.match_version = match_version
                    if len(curr_text) < const_prefix_len:
                        if pattern[len(curr_text)+1] == curr_char:
                            node.match = PARTIAL_MATCH
                        else:
                            node.match = None
                    else:
                        tmp_m = pattern_obj.match(curr_text + curr_char, partial=True)
                        node.match = tmp_m
                
                # if we have a match then we keep exploring
                if node.match:
                    if node.parent:
                        if node.parent.flag == match_version: # our parent already has another matching child, so we have multiple token possibilities
                            found_node = None
                            break
                        else:
                            node.parent.flag = match_version # mark the parent as having a matching child (us)
                            if node.value is not None:
                                found_node = node
                    
                    # set up all our children to be visited
                    curr_text += curr_char
                    node_stack.extend((curr_text, k, v) for k,v in node.children.items())

            if found_node:

                # valid_id_set = found_node.values()
                m = found_node.match
                sampled_token_ind = found_node.value
                sampled_token = self.tokens[sampled_token_ind]
                valid_id_set = None # clear any set constraints
            
            # if there are multiple possible token matches then we compute logits and explore the token trie greedily
            else:

                # compute the order in which we prefer the tokens
                logits = self._get_logits()
                if valid_id_set is not None:
                    for id in valid_id_set:
                        logits[id] += 1000
                    valid_id_set = None
                if temperature == 0:
                    sampling_order = torch.argsort(logits, descending=True).cpu().numpy() # we need numpy so the enumerate below does not get really slow...
                else:
                    assert top_p == 1, "Still need to add support for top_p!"
                    probs = torch.nn.functional.softmax(logits / temperature, dim=-1)
                    sampling_order = torch.multinomial(probs, len(probs)).cpu().numpy()

                # find the best allowed token
                #call_count[0] = 0
                self._token_trie.match_version += 1
                gen_len = len(generated_text) + len(delayed_text)
                for i,sampled_token_ind in enumerate(sampling_order):
                    sampled_token = self.tokens[sampled_token_ind]

                    # check to see if the sampled token is allowed (TODO: consider if this needs more optimized...python loops are slow)
                    token_pos = 1
                    node = self._token_trie.children[sampled_token[0]]
                    while True:
                        if node.match_version < self._token_trie.match_version:
                            node.match_version = self._token_trie.match_version
                            if gen_len + token_pos <= const_prefix_len:
                                if pattern[gen_len + token_pos] == sampled_token[token_pos-1]:
                                    node.match = PARTIAL_MATCH
                                else:
                                    node.match = None
                            else:
                                #call_count[0] += 1
                                m = pattern_obj.match(generated_text+delayed_text+sampled_token[:token_pos], partial=True)
                                node.match = m
                        
                        if token_pos == len(sampled_token):
                            m = node.match
                            break
                        else:
                            if node.match:
                                token_pos += 1
                                node = node.children[sampled_token[token_pos-1]]
                            else:
                                m = None
                                break
                    
                    if m is not None:
                        break
                assert m is not None, f"There were no tokens found that could encode: `{pattern[gen_len + token_pos]}`, perhaps the model vocabulary does not contain this token?"
            #print("call_count", call_count[0], "`" + sampled_token + "`", i)
            
            # delay emitting if we might be starting the stop pattern
            new_text = delayed_text + sampled_token
            delayed_text = ""
            stop_match = extracted_stop_pattern.search(generated_text + new_text, partial=True)
            if stop_match and stop_match.end() - stop_match.start() > 0:

                # emit delayed text before the match start
                if stop_match.start() > len(generated_text):
                    offset = stop_match.start() - len(generated_text)
                    delayed_text = new_text[offset:]
                    new_text = new_text[:offset]
                else:
                    delayed_text = new_text
                    new_text = ""

            # if we have a full match we are done
            if not m.partial:# and len(m.groupdict()["stop"]) > 0:
                new_text += delayed_text
                generated_text += new_text
                
                # strip anything after the stop group
                chars_after_stop = len(generated_text) - m.span('stop')[0]
                if chars_after_stop > 0:
                    generated_text = generated_text[:-chars_after_stop]

                # if we exactly match the end of the pattern then we can commit to this last token 
                if m.span()[1] == len(generated_text):
                    self._cache_state["new_token_ids"].append(sampled_token_ind)
                
                if hidden_count < len(new_text) - chars_after_stop:
                    yield new_text[hidden_count:len(new_text) - chars_after_stop], m.groupdict()
                break # we are done!
            else:
                generated_text += new_text

                # yeild the snippet of text created by the next token
                out = new_text[hidden_count:]
                if len(out) > 0:
                    yield out, m.groupdict()
                    hidden_count = 0
                else:
                    hidden_count -= len(new_text)

                self._cache_state["new_token_ids"].append(sampled_token_ind)


import re
import os

from ._model import Model, Chat
from guidance._grammar import Select, Join, Byte


chat_model_pattern = r'^(gpt-3\.5-turbo|gpt-4)(-\d+k)?(-\d{4})?$'

class OpenAI(Model):
    def __init__(self, model, api_key=None, organization=None, base_url=None, echo=True, eos_token="<|endoftext|>"):
        '''Initialize an OpenAI model. Parameters are based on the new OpenAI V1 SDK.

        Args:
            model: Supported OpenAI model name.
            api_key: OpenAI API key. Defaults to OPENAI_API_KEY environment variable.
            organization: OpenAI organization ID. Defaults to OPENAI_ORG_ID environment variable.
            base_url: OpenAI API base URL. Defaults to "https://api.openai.com/v1".
            echo: Whether to echo the prompt back to the output. Defaults to True.
        '''
        # TODO: Re-enable caching, retry, throttling support.

        # subclass to OpenAIChat if model is chat
        if re.match(chat_model_pattern, model) and self.__class__ is OpenAI:
            # Will enable/test Chat support after getting basic completion models working.
            raise Exception("OpenAI chat models are not yet supported. Use gpt-3.5-instruct-turbo for now.") 
            self.__class__ = OpenAIChat
            OpenAIChat.__init__(self, model=model, caching=caching)
            return
        
        # standard init
        super().__init__(model)
        self.model_name = model

        # Configure an AsyncOpenAI Client with user params.
        if api_key is None:
            api_key = os.environ.get("OPENAI_API_KEY")

        if organization is None:
            organization = os.environ.get("OPENAI_ORG_ID")

        if base_url is None:
            base_url = r"https://api.openai.com/v1"

        try:
            from openai import OpenAI as OpenAIClient # Need to avoid naming conflict
            import tiktoken
        except ImportError:
            raise ImportError("OpenAI support requires openai >= 1 and tiktoken to be installed.")
        
        self.tokenizer = tiktoken.encoding_for_model(model)
        if isinstance(eos_token, str):
            self.eos_token = bytes(eos_token, "utf-8")
        else:
            self.eos_token = b"<|endoftext|>"
        
        eos_token_id = self.tokenizer.encode(self.eos_token.decode(), allowed_special={self.eos_token.decode()})
        assert len(eos_token_id) == 1 and isinstance(eos_token_id[0], int), "Could not find a single token representation of EOS token."
        self.eos_token_id = eos_token_id[0]


        self.client = OpenAIClient(api_key=api_key, organization=organization, base_url=base_url)

    def _openai_completion_call(self, prompt, max_tokens=100, n=1, top_p=1, temperature=0.0):
        try:
            return self.client.completions.create(
                model=self.model_name, 
                prompt=prompt, 
                max_tokens=max_tokens, 
                n=n, 
                top_p=top_p, 
                temperature=temperature, 
                stream=True
            )
        except Exception as e: # TODO: add retry logic
            raise e
        

    def __call__(self, grammar, max_tokens=100, n=1, top_p=1, temperature=0.0):
        # TODO: Add support for gen intermixed with generation (i.e. grammar has a fixed piece to it.

        def is_gen_grammar(grammar):
            # First, check if the root of the grammar is a Join
            if not isinstance(grammar, Join) or len(grammar.values) != 2:
                return False

            # Next, check if the first part of the Join is a Select
            zero_or_more = grammar.values[0]
            if not isinstance(zero_or_more, Select):
                return False

            # Check if one of the options in the Select is a Join that includes the Select itself
            recursion_found = any(isinstance(option, Join) and zero_or_more in option.values for option in zero_or_more.values)
            if not recursion_found:
                return False

            # Check if the second part of the Join is a Select that contains a Join
            stop_string_select = grammar.values[1]
            if not isinstance(stop_string_select, Select):
                return False

            # Check if one of the options in the Select is a Join corresponding to the stop_string
            stop_string_found = any(isinstance(option, Join) for option in stop_string_select.values)
            if not stop_string_found:
                return False

            return True
        
        def is_select_grammar(grammar):
            # Check if the grammar is of type Select
            if not isinstance(grammar, Select):
                return False

            # Check if all options within the Select are either terminal Bytes or Joins of Bytes
            for option in grammar.values:
                if isinstance(option, Byte):
                    continue
                elif isinstance(option, Join):
                    if not all(isinstance(value, Byte) for value in option.values):
                        return False
                else:
                    return False

            return True
                
        if is_gen_grammar(grammar):
            # Make API call assuming the grammar is correct right now
            for completion in self._openai_completion_call(str(self), max_tokens, n, top_p, temperature):
                
                # Check if we are done
                completion_text = completion.choices[0].text
                if completion_text == self.eos_token:
                    break # Finish if we hit our stop token -- TODO: update this to split out the stop properly from the grammar

                # Otherwise, continue to yield out bytes
                yield bytes(completion.choices[0].text, "utf-8"), True, 0.0, {}, {} # TODO: set logprobs if we have them

        elif is_select_grammar(grammar):
            # always pick the first select value just for debugging
            yield grammar.values[0].byte, False, 0.0, {}, {}
        
        # Not in a select but just want to concat text
        elif isinstance(grammar, Join) and all(isinstance(x, Byte) for x in grammar.values):
            # all terminal bytes mean we just add that straight to the text
            for byte in grammar.values:
                yield byte.byte, False, 0.0, {}, {}

        else:
            # TODO: Add logic for simple select detection and error conditions here
            pass


class OpenAIChat(OpenAI, Chat):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def tool_def(self, *args, **kwargs):
        lm = self + "<||_html:<span style='background-color: rgba(93, 63, 211, 0.15)'>_||>"
        lm = OpenAI.tool_def(lm, *args, **kwargs)
        return lm + "<||_html:</span>_||>"


import re
import functools
import time
import collections
import os

try:
    import torch
except ImportError:
    pass

from ._model import Chat
from ._local import Local


class Transformers(Local):
    def __init__(self, model=None, tokenizer=None, echo=True, caching=True, temperature=0.0, device=None, **kwargs):
        
        # fill in default model value
        if model is None:
            model = os.environ.get("TRANSFORMERS_MODEL", None)
        if model is None:
            try:
                with open(os.path.expanduser('~/.transformers_model'), 'r') as file:
                    model = file.read().replace('\n', '')
            except:
                pass

        self.model_obj, self._orig_tokenizer = self._model_and_tokenizer(model, tokenizer, **kwargs)

        if not isinstance(model, str):
            self.model = model.__class__.__name__
        self.caching = caching
        # self.current_time = time.time()
        # self.call_history = collections.deque()
        self.temperature = temperature
        if device is not None: # set the device if requested
            self.model_obj = self.model_obj.to(device)
        self.device = self.model_obj.device # otherwise note the current device

        # build the token set and pass that to the Local constructor
        # note that we prefix the tokens with a letter so some sentence peice tokenizers don't strip leading spaces.
        tkz = self._orig_tokenizer
        super().__init__(
            [tkz.convert_tokens_to_string(['a', tkz.convert_ids_to_tokens(i)])[1:] for i in range(tkz.vocab_size)],
            tkz.bos_token_id,
            tkz.eos_token_id,
            echo=echo
        )

        self._cache_state["past_key_values"] = None
        self._cache_state["logits"] = None

    def _model_and_tokenizer(self, model, tokenizer, **kwargs):

        # intantiate the model and tokenizer if needed
        if isinstance(model, str):

            # make sure transformers is installed
            try:
                import transformers
            except:
                raise Exception("Please install transformers with `pip install transformers` in order to use guidance.llms.Transformers!")

            if tokenizer is None:
                tokenizer = transformers.AutoTokenizer.from_pretrained(model, **kwargs)
            model = transformers.AutoModelForCausalLM.from_pretrained(model, **kwargs)
        
        assert tokenizer is not None, "You must give a tokenizer object when you provide a model object (as opposed to just a model name)!"

        # discover how the model handles leading spaces
        # tokens = tokenizer.encode("alpha ruby")
        # raw_coded = ''.join([tokenizer.convert_ids_to_tokens(id) for id in tokens])
        # recoded = tokenizer.decode(tokens)
        # assert len(raw_coded) == len(recoded), "The tokenizer is changing the length of the string, so you need make a special subclass to handle this model!"
        # self.leading_space_token = raw_coded[-5]
            
        return model, tokenizer

    def _get_logits(self):
        '''Computes the logits for the given token state.
        
        This overrides a method from the LocalEngine class that is used to get
        inference results from the model.
        '''

        cache_token_ids = self._cache_state["cache_token_ids"]
        new_token_ids = self._cache_state["new_token_ids"]

        # get the number of cache positions we are using
        past_key_values = self._cache_state["past_key_values"]
        past_length = past_key_values[0][0].size(-2) if past_key_values is not None else 0
        if past_length > len(cache_token_ids):
            past_length = len(cache_token_ids)-1
            self._cache_state["past_key_values"] = tuple(tuple(p[..., :past_length, :] for p in v) for v in past_key_values)

            # note we recompute the last token because we don't bother to handle the special case of just computing logits
            new_token_ids.insert(0, cache_token_ids[-1])
            cache_token_ids.pop()

        # call the model
        if len(new_token_ids) > 0:
            model_out = self.model_obj(
                input_ids=torch.tensor(new_token_ids).unsqueeze(0).to(self.device),
                past_key_values=self._cache_state["past_key_values"],
                use_cache=True,
                position_ids=torch.arange(past_length, past_length+len(new_token_ids)).unsqueeze(0).to(self.device),
                attention_mask=torch.ones(1, past_length + len(new_token_ids)).to(self.device)
            )

            # save the results
            self._cache_state["past_key_values"] = model_out.past_key_values
            cache_token_ids.extend(new_token_ids)
            self._cache_state["new_token_ids"].clear()
            self._cache_state["logits"] = model_out.logits[0, -1, :]
        
        return self._cache_state["logits"]
    
    # def __init__(self, engine, caching=True, **engine_kwargs):
    #     super().__init__(engine, caching=caching)
    #     # self.engine = engine

    #     if isinstance(self.engine, str):
    #         self.engine = guidance.endpoints.Transformers(engine, **engine_kwargs)
    #     # self._endpoint_session = self.endpoint.session()

class TransformersChat(Transformers, Chat):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)